# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

requires_core = [
    'Twisted[tls]==18.9.0',
    'pyOpenSSL>=16.2.0',
]

if sys.platform == 'darwin':
    requires_core = [
        'Twisted==18.9.0',
    ]

requires = requires_core + [

    # Core
    #'Twisted[tls]==17.1.0',
    #'Twisted==17.1.0',
    'six==1.14.0',
    'pyramid==1.5.7',               # 1.9.1
    'pyramid_jinja2==2.5',          # 2.7
    'cornice==1.0.0',               # 2.4.0, 3.1.0
    'simplejson==3.8.2',            # 3.13.2
    'Jinja2==2.8',                  # 2.10
    'Bunch==1.0.1',
    'appdirs==1.4.3',
    'json-store==2.1',
    'python-dateutil>=2.6.0',       # 2.7.0
    'arrow==0.10.0',                # 0.12.1
    'funcy==1.7.2',                 # 1.10.1
    'delegator.py==0.1.0',
    'attrs==18.2.0',

    # Bus adapters
    #'twisted-mqtt==0.1.4',         # 0.2.1
    'paho-mqtt==1.2',               # 1.4.0
    'autobahn[twisted]==0.13.0',    # 18.3.1
    'msgpack-python==0.4.7',        # 0.5.6

    # Misc
    #'setuptools==40.8.0',
    'distlib==0.2.8',
    'docopt==0.6.2',

    # More dependencies
    'cryptography>=1.3.4',
    'certifi>=2016.2.28',

    #'pyasn1>=0.2.3',                # 0.4.2
    'service_identity==18.1.0',
    'idna==2.8',
    'pyasn1==0.4.5',
    'pyasn1-modules==0.2.4',

    'pycparser==2.17',              # 2.18
    'pyparsing==2.2.0',

]

extras = {
    'daq': [
        'influxdb==4.0.0',          # 5.0.0
        'pytz==2018.9',
        'requests>=2.12.4',
        #'requests>=2.18.4,<3',
        'grafana_api_client==0.2.0',
        #'grafana-dashboard-builder==0.1.0a7',      # evaluated, but not suitable
        #'txmongo==16.3.0',
        'pymongo==3.4.0',           # 3.6.1
    ],
    'daq_geospatial': [
        'Geohash==1.0',
        'geopy==1.12.0',            # 1.17.0
        'Beaker==1.9.0',            # 1.10.0
        'tqdm==4.19.8',             # 4.28.1
    ],
    'daq_binary': [
        'pyclibrary==0.1.3',
        'tabulate==0.7.5',          # 0.8.2
        'sympy==0.7.6.1',           # 1.1.1
    ],
    'storage_plus': [
        'alchimia==0.4',            # 0.8.1
    ],

    # Data export: Basic formats
    'export': [
        'pyinfluxql==0.0.1',
        'pandas==0.18.1',           # 0.24.1
        'numpy>=1.8.2,<2',          # 1.14.2
        'XlsxWriter==0.9.2',        # 1.0.2
    ],

    'plotting': [
        #'dyplot==0.8.8',

        # sudo port install py27-matplotlib
        'matplotlib==2.0.0',        # 2.2.2
        #'cairocffi>=0.5.4',
        'bokeh==0.12.4',            # 0.12.14
        'vincent==0.4.4',
    ],

    # Data export: Scientific data formats like HDF5 and NetCDF and plots from ggplot
    'scientific': [

        # Data
        # ----
        # HDF5
        # "PyTables" requires HDF5 libraries
        # sudo port install hdf5
        'tables>=3.1.1',            # 3.4.2

        # NetCDF (Network Common Data Form)
        'xarray==0.7.2',            # 0.10.2
        # sudo port install netcdf
        'netCDF4>=1.2.7',           # 1.3.1
        #'h5netcdf==0.2.2',

        # Algorithms
        # ----------
        # sudo port install py27-scipy
        'scipy>=0.14.0',            # 1.0.0
        'ggplot==0.9.7',            # 0.11.5

        # gfortran
        # aptitude install libatlas-base-dev lapack-dev gfortran or
        # https://gcc.gnu.org/wiki/GFortranBinaries

        # Visualization
        # -------------
        #'seaborn==0.7.1',

    ],

    'firmware': [
        'GitPython==2.0.5',         # 2.1.8
        'plumbum==1.6.1.post0',     # 1.6.6
    ],

}

setup(name='kotori',
      version='0.24.1',
      description='Kotori is a data acquisition, processing and graphing toolkit for humans',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Framework :: Pyramid",
        "Framework :: Twisted",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Object Brokering",
        "Topic :: System :: Archiving",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: Text Processing",
        "Topic :: Utilities",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
        ],
      author='Andreas Motl',
      author_email='andreas@getkotori.org',
      url='https://github.com/daq-tools/kotori',
      keywords='data acquisition graphing export plotting daq routing engine ' +
               'mqtt http rest amqp wamp sql web html csv json cdf hdf5 png ' +
               'twisted pyramid autobahn influxdb mosquitto grafana mongodb matplotlib ggplot ' +
               'telemetry m2m iot',
      packages=find_packages(),
      include_package_data=True,
      package_data={
        'kotori': [
          'daq/graphing/grafana/resources/*.json',
          'io/export/*.html',
        ],
      },
      zip_safe=False,
      test_suite='kotori.test',
      install_requires=requires,
      extras_require=extras,
      dependency_links=[
          'https://github.com/jjmalina/pyinfluxql/tarball/d92db4ab8c#egg=pyinfluxql-0.0.1',
      ],
      entry_points={
          'console_scripts': [
              'kotori              = kotori:run',
              #'kotori-master      = kotori.master.server:run',
              #'kotori-node        = kotori.node.nodeservice:run',
              #'kotori-wamp-client = kotori.master.client:run_wamp_client',
              'h2m-csv-udp-client  = kotori.vendor.hydro2motion.client:run_udp_client',
              'h2m-csv-udp-fuzzer  = kotori.vendor.hydro2motion.client:run_udp_fuzzer',
              'lst-message         = kotori.vendor.lst.shell:message',
              'kotori-selftest     = kotori.vendor.selftest:run',
          ],
          'paste.app_factory': [
              'main = kotori.frontend.app:main',
          ],
      },
)



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
# https://pypi.python.org/pypi/twisted-mqtt
# https://github.com/astrorafael/twisted-mqtt/
from __future__ import absolute_import
from twisted.logger import Logger

log = Logger()

class BaseMqttAdapter(object):

    def __init__(self, name,
                 broker_host=u'localhost', broker_port=1883, broker_username=None, broker_password=None,
                 debug=False, callback=None, subscriptions=None):
        self.name = name
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.broker_username = broker_username
        self.broker_password = broker_password
        self.callback = callback or self.on_message
        self.subscriptions = subscriptions or []

    def startService(self):
        self.log(log.info, u'Starting')
        self.running = 1
        self.connect()
        #reactor.callInThread(self.connect)

    def connect(self):
        raise NotImplementedError('Please implement the "connect" method in derived class')

    def subscribe(self, *args):
        raise NotImplementedError('Please implement the "subscribe" method in derived class')

    def on_message(self, topic=None, payload=None, **kwargs):
        # former def on_message(self, topic, payload, qos, dup, retain, msgId):
        # kwargs: qos, dup, retain, msgId
        raise NotImplementedError('Please implement the "on_message" method in derived class or pass a callable using the "callback" constructor argument')

    def on_error(self, *args):
        log.error("ERROR: args={args!s}", args=args)

    def log(self, callable, prefix):
        callable(u'{prefix} {class_name}. name={name}, broker={broker_host}:{broker_port}, object={repr}',
            prefix=prefix, class_name=self.__class__.__name__,
            name=self.name, repr=repr(self),
            broker_host=self.broker_host, broker_port=self.broker_port)

    """
    def _todo_publish(self):
        d = self.protocol.publish(topic="foo/bar/baz", message="hello friends")
        d.addErrback(self.on_error)

    def _todo_prepareToPublish(self, *args):
        self.task = task.LoopingCall(self.publish)
        self.task.start(5.0)
    """



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)

class BinaryMessageAdapter(object):

    def __init__(self, struct_registry):
        self.struct_registry = struct_registry

    def decode(self, payload):

        # decode struct ID from binary data, it's always at the second byte
        message_id = ord(payload[1])
        logger.debug('Message struct ID is "{}"'.format(message_id))

        # look up proper StructAdapter object
        struct_schema = self.struct_registry.get_by_id(message_id)

        # create instance of struct
        struct = struct_schema.create()

        # load binary message payload into struct
        struct._load_(payload)

        return struct

    def pprint(self, struct, data=None, format='tabulate-plain'):
        # pretty-print struct content
        # TODO: maybe refactor to struct._pprint_
        self.struct_registry.pprint(struct, data=data, format=format)

    def to_dict(self, struct):
        # convert struct content to json
        # TODO: maybe refactor to struct._json_
        return self.struct_registry.to_dict(struct)

    def transform(self, struct):
        """
        apply transformation rules to struct
        """

        # convert struct to OrderedDict
        data = self.struct_registry.to_dict(struct)

        # look up proper StructAdapter object
        struct_schema = self.struct_registry.get_by_id(struct.ID)

        # get annotations object
        annotations = struct_schema.annotations

        # prepare variables suitable for expression evaluation
        # TODO: make check against "null bytes" more generic
        variables = dict(data)
        del variables['ck']     # otherwise: "TypeError: expected string without null bytes" from SymPy

        # apply transformation rules, keep item order
        data_new = OrderedDict()
        for key, value in data.iteritems():

            # get transformation rule
            rule = annotations.get_struct_rule(struct_schema.name, key)
            if rule:

                # apply expression
                if 'expression_sympy' in rule:
                    expr = rule['expression_sympy']

                    # always gives (SymPy) Floats or Zero, never Integer :-(
                    value = expr.subs(variables).evalf()

                    # convert to Python-native data types
                    # TODO: check if SymPy provides a generic builtin for that
                    if value.is_zero:
                        value = 0.0
                    elif value.is_Float:
                        value = float(value)

                # apply field renaming
                if 'name' in rule:
                    key = rule['name']

            data_new[key] = value

        return data_new




# -*- coding: utf-8 -*-
# (c) 2014 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from string import Template
from pkg_resources import resource_filename
from twisted.logger import Logger
from twisted.internet import reactor
from twisted.web.resource import Resource
from twisted.web.server import Site
from twisted.web.static import File
from kotori.util.common import NodeId, get_hostname

log = Logger()


class CustomTemplate(Template):
    delimiter = '$$'

class WebDashboard(Resource):
    def getChild(self, name, request):
        print 'getChild:', name
        if name == '':
            return WebDashboardIndex()
        else:
            document_root = resource_filename('kotori.web', '')
            return File(document_root)

class WebDashboardIndex(Resource):

    def __init__(self, websocket_uri, filename):
        Resource.__init__(self)
        self.websocket_uri = websocket_uri
        self.filename = filename

    def render_GET(self, request):
        index = resource_filename('kotori.vendor.hydro2motion.web', self.filename)
        tpl = CustomTemplate(file(index).read())
        response = tpl.substitute({
            'websocket_uri': self.websocket_uri,
            'node_id': str(NodeId()),
            'hostname': get_hostname(),
        })
        return response.encode('utf-8')

def boot_web(settings, debug=False):

    http_port = int(settings.hydro2motion.http_port)
    websocket_uri = unicode(settings.wamp.uri)

    dashboard = Resource()
    dashboard.putChild('', WebDashboardIndex(websocket_uri=websocket_uri, filename='index.html'))
    dashboard.putChild('fs.html', WebDashboardIndex(websocket_uri=websocket_uri, filename='fs.html'))
    dashboard.putChild('poly.html', WebDashboardIndex(websocket_uri=websocket_uri, filename='poly.html'))
    dashboard.putChild('static', File(resource_filename('kotori.vendor.hydro2motion.web', 'static')))

    log.info('Starting HTTP service on port {http_port}', http_port=http_port)
    factory = Site(dashboard)
    reactor.listenTCP(http_port, factory)



# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
import re
from copy import deepcopy
from urlparse import urlparse
from bunch import Bunch, bunchify
from twisted.logger import Logger
from twisted.application.service import MultiService
from kotori.util.configuration import read_list
from kotori.core import KotoriBootloader
from kotori.daq.services import RootService, MultiServiceMixin
from kotori.io.protocol.http import HttpServerService
from kotori.io.protocol.target import ForwarderTargetService

log = Logger()

class ProtocolForwarderApplication(RootService):
    """
    Application service container root object
    """

    def __init__(self, name=None, application_settings=None, global_settings=None):
        RootService.__init__(self, settings=global_settings)

        # Compute name for Twisted service
        self.name = u'app-{name}'.format(name=name)

        # Make channel object from application settings configuration object
        self.setupChannel(channel=application_settings, name=name)

        # Create application service object composed of subsystem components
        service = ProtocolForwarderService(
            channel = self.channel
        )

        # Register service component with its container
        self.registerService(service)


class ProtocolForwarderService(MultiServiceMixin, MultiService):
    """
    Container service for source and target services.

    Both services are connected unidirectionally via ``forward`` callback.

    As of June 2016, the only source service is a singleton
    instance of a ``HttpServerService``. This might be
    improved by adding more handlers.

    Also, there are currently two target services
    for emitting data, MQTT and InfluxDB.
    """

    def __init__(self, channel=None):
        self.channel = channel or Bunch(realm=None, subscriptions=[])
        MultiServiceMixin.__init__(self, name=self.channel.name + '-forwarder')

    def setupService(self):
        #self.log(log.info, u'Setting up')
        log.info(u'Starting {name}'.format(name=self.logname))

        self.settings = self.parent.settings

        # Configure metrics to be collected each X seconds
        #self.metrics = Bunch(tx_count=0, starttime=time.time(), interval=2)

        log.info('Forwarding payloads from {source} to {target}', **self.channel)
        self.setupSource()
        self.setupTarget()

    def setupSource(self):
        """
        Configure data source by registering a HTTP endpoint
        using the source address derived from channel settings.
        """

        # There should be just a single instance of a HTTP server service object
        self.source_service = HttpServerService.create(self.settings)

        # Wrap source channel settings into address object
        self.source_address = ForwarderAddress(self.channel.source)

        # Register URI route representing source channel
        self.source_service.registerEndpoint(
            methods     = self.source_address.predicates,
            path        = self.source_address.uri.path,
            callback    = self.forward)

    def setupTarget(self):
        """
        Configure data target by registering a service object
        using the target address derived from channel settings.
        """

        # Wrap target channel into address object
        self.target_address = ForwarderAddress(self.channel.target)
        self.target_uri = self.target_address.uri

        # Compute name for service object, should be unique.
        target_service_name = u'{realm}-{channel_name}'.format(
            realm=self.channel.realm, channel_name=self.channel.name)

        # Register service representing target channel.
        # Each service should just run once, so they are named to be found again.
        # TODO: Add sanity checks for collisions here.
        try:
            self.target_service = self.getServiceNamed(target_service_name)
        except KeyError:
            self.target_service = ForwarderTargetService(name=target_service_name, address=self.target_address)
            self.registerService(self.target_service)

    def forward(self, bucket):
        """
        Receive data bucket from source, run through
        transformation machinery and emit to target.
        """

        # 1. Map/transform topology address information
        if 'transform' in self.channel:
            for entrypoint in read_list(self.channel.transform):
                try:
                    transformer = KotoriBootloader.load_entrypoint(entrypoint)
                    bucket.tdata.update(transformer(bucket.tdata))
                except ImportError as ex:
                    log.error('ImportError "{message}" when loading entrypoint "{entrypoint}"',
                        entrypoint=entrypoint, message=ex)

        # MQTT doesn't prefer leading forward slashes with topic names, let's get rid of them
        target_uri_tpl = self.target_uri.path.lstrip('/')

        # Compute target bus topic from url matches
        target_uri = target_uri_tpl.format(**bucket.tdata)

        # Enrich bucket by putting source and target addresses into it
        bucket.address = Bunch(source=self.source_address, target=self.target_address)

        # 2. Reporting
        bucket_logging = Bunch(bucket)
        if 'body' in bucket_logging and len(bucket_logging.body) > 100:
            bucket_logging.body = bucket_logging.body[:100] + ' [...]'
        log.debug('Forwarding bucket to {target} with bucket={bucket}. Effective target uri is {target_uri}',
            target=self.channel.target, target_uri=target_uri, bucket=dict(bucket_logging))

        # 3. Adapt, serialize and emit appropriately
        return self.target_service.emit(target_uri, bucket)

    def log(self, level, prefix):
        level('{prefix} {class_name}. name={name}, channel={channel}',
            prefix=prefix, class_name=self.__class__.__name__, name=self.name, channel=dict(self.channel))


class ForwarderAddress(object):
    """
    Addresses are made of uris and predicates, e.g.::

        http:/api/mqttkit-1/{address:.*}/data [POST]
        ^                                      ^
        |                                      |
        ----- uri                   predicate --

    Synopsis:

    >>> address = ForwarderAddress(u'http:/api/mqttkit-1/{address:.*}/data [POST]')

    >>> address.uri.scheme
    u'http'

    >>> address.uri.path
    u'/api/mqttkit-1/{address:.*}/data'

    >>> address.predicates
    [u'POST']
    """

    address_pattern = re.compile('^(?P<uri>.*?)(?: \[(?P<predicates>.+)\])?$')

    def __init__(self, address):
        self.raw_uri = None
        self.raw_predicates = None
        self.parsed_uri = None
        self.uri = None
        self.predicates = []
        self.parse(address)

    def parse(self, address):
        m = self.address_pattern.match(address)

        match = m.groupdict()
        if match:
            self.raw_uri = match['uri']
            self.raw_predicates = match['predicates']

            self.parsed_uri = urlparse(self.raw_uri)
            self.uri = bunchify(self.parsed_uri._asdict())
            if self.raw_predicates:
                self.predicates = read_list(self.raw_predicates)

    def __repr__(self):
        return u'{uri} {predicates}'.format(uri=self.parsed_uri.geturl(), predicates=self.predicates)


def boot(name=None, **kwargs):
    app = ProtocolForwarderApplication(name=name, **kwargs)
    return app




# -*- coding: utf-8 -*-
"""
Kotori is a data acquisition, routing and graphing toolkit
Copyright (C) 2014-2019 Andreas Motl, <andreas@getkotori.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from pkg_resources import EntryPoint
from pyramid.settings import asbool
from twisted.logger import Logger, LogLevel
from kotori.util.configuration import read_list
from kotori.util.errors import last_error_and_traceback
from kotori.util.logger import changeLogLevel
from kotori.version import __VERSION__

log = Logger()
APP_NAME = u'Kotori version ' + __VERSION__

# TODO: introduce kotori.core.boot_component(s)

class KotoriBootloader(object):

    def __init__(self, settings=None):
        self.settings = settings

    def boot_applications(self):
        """
        Boot all enabled applications
        """
        applications = list(self.get_applications())
        log.info(u'Enabling applications {applications}', applications=applications)
        for name in applications:
            self.boot_application(name)

    def boot_application(self, name):
        """
        Boot application defined by configuration
        """
        log.info(u'Starting application "{name}"', name=name)

        # Initialize application and run as Twisted service
        factory = self.get_application_factory(name)

        if not factory:
            return

        try:
            application_settings = self.settings[name]
            application = factory(name=name, application_settings=application_settings, global_settings=self.settings)
            application.startService()
        except Exception as ex:
            log.failure('Error running application factory "{name}":\n{log_failure}', name=name)
            return

    def get_application_factory(self, name):

        # Get application information from configuration object
        try:
            application_settings = self.settings[name]
            app_factory = \
                'app_factory' in application_settings and \
                application_settings.app_factory or \
                application_settings.application
        except:
            log.failure('Application configuration object for "{name}" not found, missing setting "app_factory"', name=name)
            return

        try:
            factory_callable = self.load_entrypoint(app_factory)
        except:
            log.failure('Error loading application entrypoint "{app_factory}" for "{name}":\n{ex}',
                name=name, app_factory=app_factory, ex=last_error_and_traceback())
            return

        return factory_callable

    @classmethod
    def load_entrypoint(cls, reference, onerror='raise'):

        # Resolve entrypoint
        expression = u'_ = {reference}'.format(reference=reference)
        try:
            entrypoint = EntryPoint.parse(expression)
        except:
            log.failure('Error parsing entrypoint "{reference}" from expression "{expression}"',
                reference=reference, expression=expression)
            raise

        # Load entrypoint
        try:
            thing = entrypoint.load(require=False)
        except:
            log.failure('Error loading entrypoint "{reference}"', reference=reference)
            if onerror == 'raise':
                raise
            elif onerror == 'ignore':
                return cls.noop_callable
            #raise

        return thing

    @classmethod
    def noop_callable(cls, *args, **kwargs):
        pass

    def get_vendors(self):
        for name, config_object in self.settings.iteritems():
            if 'type' in config_object and config_object.type == 'vendor':
                if 'enable' not in config_object or asbool(config_object['enable']):
                    yield name

    def get_applications(self):
        for name, config_object in self.settings.iteritems():
            if 'type' in config_object and config_object.type == 'application':
                if 'enable' not in config_object or asbool(config_object['enable']):
                    yield name

    def boot_vendors(self):
        """
        Boot all enabled vendors.
        """
        vendors = list(self.get_vendors())
        log.info('Enabling vendors {vendors}', vendors=vendors)

        debug = self.settings.options.debug

        for vendor in vendors:
            if hasattr(VendorBootloader, vendor):
                log.info(u'Starting vendor environment for "{vendor}"', vendor=vendor)
                bootloader = getattr(VendorBootloader, vendor)
                try:
                    bootloader(self.settings)
                except Exception as ex:
                    log.failure(
                        'Error booting vendor environment for "{vendor}"":\n{log_failure}"', vendor=vendor)
            else:
                message = 'Vendor environment for "{vendor}" missing.'.format(vendor=vendor)
                log.critical(message)


class VendorBootloader(object):

    @classmethod
    def hydro2motion(cls, settings):
        from kotori.vendor.hydro2motion.database.influx import h2m_boot_influx_database
        from kotori.vendor.hydro2motion.network.udp import h2m_boot_udp_adapter
        from kotori.vendor.hydro2motion.web.server import boot_web

        debug = settings.options.debug
        boot_web(settings, debug=debug)
        h2m_boot_udp_adapter(settings, debug=debug)
        h2m_boot_influx_database(settings)

    @classmethod
    def hiveeyes(cls, settings):
        from kotori.vendor.hiveeyes.application import hiveeyes_boot
        debug = settings.options.debug
        hiveeyes_boot(settings, debug=debug)

    @classmethod
    def lst(cls, settings):
        from kotori.vendor.lst.application import lst_boot
        debug = settings.options.debug
        if settings.options.debug_vendor and 'lst' in read_list(settings.options.debug_vendor):
            changeLogLevel('kotori.vendor.lst', LogLevel.debug)
        lst_boot(settings)



# -*- coding: utf-8 -*-
from pyclibrary.c_parser import CParser
from pyclibrary.thirdparty.pyparsing import (Forward, Optional, Keyword, Literal, ZeroOrMore, OneOrMore, Group, nestedExpr, delimitedList, Suppress)

# taken verbatim from pyclibrary.c_parser
# 2015-11-07 make pyclibrary grok more constructor flavours

class CParserEnhanced(CParser):

    def build_parser(self):
        """Builds the entire tree of parser elements for the C language (the
        bits we support, anyway).
        """

        from pyclibrary.c_parser import \
            fund_type, kwl, size_modifiers, sign_modifiers, extra_modifier, recombine, type_qualifier, ident, \
            lparen, rparen, expression, colon, semi, integer, storage_class_spec, rbrace, lbrace, number, comma, \
            lbrack, rbrack, call_conv

        if hasattr(self, 'parser'):
            return self.parser

        self.struct_type = Forward()
        self.enum_type = Forward()
        type_ = (fund_type |
                 Optional(kwl(size_modifiers + sign_modifiers)) + ident |
                 self.struct_type |
                 self.enum_type)
        if extra_modifier is not None:
            type_ += extra_modifier
        type_.setParseAction(recombine)
        self.type_spec = (type_qualifier('pre_qual') +
                          type_("name"))

        # --- Abstract declarators for use in function pointer arguments
        #   Thus begins the extremely hairy business of parsing C declarators.
        #   Whomever decided this was a reasonable syntax should probably never
        #   breed.
        #   The following parsers combined with the process_declarator function
        #   allow us to turn a nest of type modifiers into a correctly
        #   ordered list of modifiers.

        self.declarator = Forward()
        self.abstract_declarator = Forward()

        #  Abstract declarators look like:
        #     <empty string>
        #     *
        #     **[num]
        #     (*)(int, int)
        #     *( )(int, int)[10]
        #     ...etc...
        self.abstract_declarator << Group(
            type_qualifier('first_typequal') +
            Group(ZeroOrMore(Group(Suppress('*') + type_qualifier)))('ptrs') +
            ((Optional('&')('ref')) |
             (lparen + self.abstract_declarator + rparen)('center')) +
            Optional(lparen +
                     Optional(delimitedList(Group(
                         self.type_spec('type') +
                         self.abstract_declarator('decl') +
                         Optional(Literal('=').suppress() + expression,
                             default=None)('val')
                     )), default=None) +
                     rparen)('args') +
            Group(ZeroOrMore(lbrack + Optional(expression, default='-1') +
                             rbrack))('arrays')
        )

        # Declarators look like:
        #     varName
        #     *varName
        #     **varName[num]
        #     (*fnName)(int, int)
        #     * fnName(int arg1=0)[10]
        #     ...etc...
        self.declarator << Group(
            type_qualifier('first_typequal') + call_conv +
            Group(ZeroOrMore(Group(Suppress('*') + type_qualifier)))('ptrs') +
            ((Optional('&')('ref') + ident('name')) |
             (lparen + self.declarator + rparen)('center')) +
            Optional(lparen +
                     Optional(delimitedList(
                         Group(self.type_spec('type') +
                               (self.declarator |
                                self.abstract_declarator)('decl') +
                               Optional(Literal('=').suppress() +
                                        expression, default=None)('val')
                         )),
                         default=None) +
                     rparen)('args') +
            Group(ZeroOrMore(lbrack + Optional(expression, default='-1') +
                             rbrack))('arrays')
        )
        self.declarator_list = Group(delimitedList(self.declarator))

        # Typedef
        self.type_decl = (Keyword('typedef') + self.type_spec('type') +
                          self.declarator_list('decl_list') + semi)
        self.type_decl.setParseAction(self.process_typedef)

        # Variable declaration
        self.variable_decl = (
            Group(storage_class_spec +
                  self.type_spec('type') +
                  Optional(self.declarator_list('decl_list')) +
                  Optional(Literal('=').suppress() +
                           (expression('value') |
                            (lbrace +
                             Group(delimitedList(expression))('array_values') +
                             rbrace
                                )
                               )
                  )
            ) +
            semi)
        self.variable_decl.setParseAction(self.process_variable)

        # Function definition
        self.typeless_function_decl = (self.declarator('decl') +
                                       nestedExpr('{', '}').suppress())
        self.function_decl = (storage_class_spec +
                              self.type_spec('type') +
                              self.declarator('decl') +
                              nestedExpr('{', '}').suppress())
        self.function_decl.setParseAction(self.process_function)

        # Struct definition
        self.struct_decl = Forward()
        struct_kw = (Keyword('struct') | Keyword('union'))
        self.struct_member = (
            Group(self.variable_decl.copy().setParseAction(lambda: None)) |
            # Hack to handle bit width specification.
            Group(Group(self.type_spec('type') +
                        Optional(self.declarator_list('decl_list')) +
                        colon + integer('bit') + semi)) |
            (self.type_spec + self.declarator +
             nestedExpr('{', '}')).suppress() |
            (self.declarator + nestedExpr('{', '}')).suppress()
            # 2015-11-07 make pyclibrary grok more constructor flavours::
            #
            #    struct_program()
            #    : length(13), ID(0)
            #    {}
            | (self.declarator + ':' + OneOrMore(ident + '(' + OneOrMore(number) + ')' + Optional(',')) + nestedExpr('{', '}')).suppress()
            )

        self.decl_list = (lbrace +
                          Group(OneOrMore(self.struct_member))('members') +
                          rbrace)
        self.struct_type << (struct_kw('struct_type') +
                             ((Optional(ident)('name') +
                               self.decl_list) | ident('name'))
            )
        self.struct_type.setParseAction(self.process_struct)

        self.struct_decl = self.struct_type + semi

        # Enum definition
        enum_var_decl = Group(ident('name') +
                              Optional(Literal('=').suppress() +
                                       (integer('value') | ident('valueName'))))

        self.enum_type << (Keyword('enum') +
                           (Optional(ident)('name') +
                            lbrace +
                            Group(delimitedList(enum_var_decl))('members') +
                            Optional(comma) + rbrace | ident('name'))
            )
        self.enum_type.setParseAction(self.process_enum)
        self.enum_decl = self.enum_type + semi

        self.parser = (self.type_decl | self.variable_decl |
                       self.function_decl)
        return self.parser



# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
import re
import json
import types
import mimetypes
import pymongo
from six import BytesIO
from copy import deepcopy
from urlparse import urlparse
from bunch import bunchify, Bunch
from collections import OrderedDict
from twisted.application.service import Service
from twisted.internet import reactor, threads
from twisted.logger import Logger
from twisted.web import http, server
from twisted.web.http import parse_qs
from twisted.web.resource import Resource
from twisted.web.server import Site
from twisted.web.error import Error
from twisted.python.compat import nativeString
from kotori.io.router.path import PathRoutingEngine
from kotori.io.export.tabular import UniversalTabularExporter
from kotori.io.export.plot import UniversalPlotter
from kotori.io.protocol.util import convert_floats, slugify_datettime, flatten_request_args, handleFailure
from kotori.util.errors import last_error_and_traceback

log = Logger()


class LocalSite(Site):

    def log(self, request):
        """
        Redirect logging of HTTPFactory.

        @param request: The request object about which to log.
        @type request: L{Request}
        """
        line = u'HTTP access: ' + self._logFormatter(self._logDateTime, request)
        if hasattr(self, '_nativeize') and self._nativeize:
            line = nativeString(line)
        else:
            line = line.encode("utf-8")
        log.debug(line)


class HttpServerService(Service):
    """
    Singleton instance of a Twisted service wrapping
    the Twisted TCP/HTTP server object "Site", in turn
    obtaining a ``HttpChannelContainer`` as root resource.
    """

    _instance = None

    def __init__(self, settings):

        # Propagate global settings
        self.settings = settings

        # Unique name of this service
        self.name = 'http-server-default'

        # Root resource object representing a channel
        # Contains routing machinery
        self.root = HttpChannelContainer()

        # Forward route registration method to channel object
        self.registerEndpoint = self.root.registerEndpoint

    def startService(self):
        """
        Start TCP listener on designated HTTP port,
        serving ``HttpChannelContainer`` as root resource.
        """

        # Don't start service twice
        if self.running == 1:
            return

        self.running = 1

        # Prepare startup
        http_listen = self.settings.kotori.http_listen
        http_port   = int(self.settings.kotori.http_port)
        log.info('Starting HTTP service on {http_listen}:{http_port}', http_listen=http_listen, http_port=http_port)

        # Configure root Site object and start listening to requests.
        # This must take place only once - can't bind to the same port multiple times!
        factory = LocalSite(self.root)
        reactor.listenTCP(http_port, factory, interface=http_listen)

    @classmethod
    def create(cls, settings):
        """
        Singleton factory
        """
        if not cls._instance:
            cls._instance = HttpServerService(settings)
            cls._instance.startService()
        return cls._instance


class HttpChannelContainer(Resource):
    """
    Twisted Site HTTP root resource driven by a
    routing engine based on the Pyramid request router.
    """

    def __init__(self):
        Resource.__init__(self)

        log.info('Initializing HttpChannelContainer')

        # Reference to Metastore database.
        self.metastore = None

        # HTTP routing callbacks.
        self.callbacks = {}

        # Connect to Metastore database.
        self.database_connect()

        # Initialize routing engine.
        self.router    = PathRoutingEngine()

    def database_connect(self):
        """
        Connect to Metadata storage
        """
        log.info('Connecting to Metadata storage database')

        # TODO: Make MongoDB address configurable
        mongodb_uri = "mongodb://localhost:27017"
        try:
            self.metastore = pymongo.MongoClient(host='localhost', port=27017, socketTimeoutMS=5000, connectTimeoutMS=5000)

        except Exception as ex:
            log.failure('Could not connect to Metadata storage database: {ex}\n{log_failure}', ex=ex)

    def registerEndpoint(self, methods=None, path=None, callback=None):
        """
        Register path/callback with routing engine.
        """
        methods = methods or []
        log.info("Registering endpoint at path '{path}' for methods {methods}", path=path, methods=methods)
        if not callable(callback):
            log.error('Reference to endpoint {path} specified via "callback" '
                      'argument is not callable: {callback}', path=path, callback=callback)
            return

        # TODO: Add sanity checks for protecting against collisions on "name" and "path"
        name = path
        log.debug('Adding route: name={name}, path={path}, methods={methods}', name=name, path=path, methods=methods)
        self.router.add_route(name, path, methods=methods)
        self.callbacks[name] = callback

    def getChild(self, name, request):
        """
        Twisted Resource path traversal method using
        the Pyramid request router for matching
        the request to registered endpoints.

        Returns ``HttpChannelEndpoint`` instance on match.
        """

        # Read URL from request.
        url_raw = str(request.URLPath())
        log.debug('HttpChannelContainer.getChild: method={method}, name={name}, url={url}',
                  method=request.method, name=name, url=url_raw)

        # Parse URL string.
        uri = urlparse(url_raw)


        # Dispatch URL to resource.

        # Router v1
        """
        for endpoint in self.endpoints:
            if uri.path.startswith(endpoint.path):
                return HttpChannelEndpoint(options=endpoint)
        """

        # Router v2
        #print 'Matching method=', request.method, 'uri=', uri.path
        result = self.router.match(request.method, uri.path)
        #print 'result:', result

        if result:
            #print 'route match result:'; from pprint import pprint; pprint(result); pprint(result['route'].__dict__)

            # Obtain matched route name.
            route_name = result['route'].name

            # Obtain appropriate callback function.
            # TODO: Beware of collisions on "route_name", see above
            callback = self.callbacks[route_name]

            # Wrap endpoint description into container object.
            endpoint = bunchify({'path': route_name, 'callback': callback, 'match': result['match'], 'request': request})
            #print 'endpoint:'; pprint(endpoint)

            # Create leaf resource instance.
            return HttpChannelEndpoint(options=endpoint, metastore=self.metastore)

        # If nothing matched, continue traversal.
        return self


class HttpChannelEndpoint(Resource):
    """
    Upper layer of data forwarding workhorse for HTTP.

    Twisted Site HTTP leaf resource containing the main dispatcher
    logic for forwarding inbound requests to the routing target.
    """

    isLeaf = True

    def __init__(self, options, metastore=None):
        self.options = options
        self.metastore = metastore
        Resource.__init__(self)

    def render(self, request):
        """
        Main Twisted Resource rendering method,
        overridden to provide custom logic.
        """

        # Pluck ``error_response`` method to request object
        request.error_response = self.error_response

        # Pluck ``channel_identifier`` attribute to request object
        request.channel_identifier = re.sub('/data.*$', '', request.path.replace('/api', ''))

        # Pluck response messages object to request object
        request.messages = []

        # Add informational headers
        request.setHeader('Channel-Id', request.channel_identifier)

        # Main bucket data container object serving the whole downstream processing chain
        bucket = Bunch(path=request.path, request=request)

        # Request processing chain, worker-threaded
        deferred = threads.deferToThread(self.dispatch, request, bucket)
        deferred.addErrback(handleFailure, request)
        deferred.addBoth(self.render_messages, request)
        deferred.addBoth(request.write)
        deferred.addBoth(lambda _: request.finish())
        return server.NOT_DONE_YET

    def dispatch(self, request, bucket):
        """
        Dispatch request
        """

        # Read and decode/parse ingress data
        data = self.read_request(bucket)

        # Process ingress data
        response = self.process_data(data, bucket)

        # Send response
        return response

    def render_messages(self, passthrough, request):
        if request.messages:
            request.setHeader('Content-Type', 'application/json')
            return json.dumps(request.messages, indent=4)
        else:
            #request.setHeader('Content-Type', 'text/plain; charset=utf-8')
            return passthrough

    def read_request(self, bucket):
        """
        Forward inbound requests to the routing target by performing these steps:

        - Build the transformation data container ``tdata``
          by feeding it information from the HTTP request:

            - Obtain HTTP request
            - Decode data from request body
            - Merge data from request arguments
            - Merge data from url matches

        - Build the main data container object ``bucket``
          serving the whole downstream processing chain.

        - Call designated registered callback method with ``bucket``.

        """

        request = bucket.request

        content_type = request.getHeader('Content-Type')
        log.debug('Received HTTP request on uri {uri}, '
                  'content type is "{content_type}"', uri=request.path, content_type=content_type)

        # Data acquisition uses HTTP POST
        if request.method == 'POST':
            return self.data_acquisition(bucket)

    def data_acquisition(self, bucket):

        request = bucket.request

        content_type = request.getHeader('Content-Type')

        # Read and decode request body
        body = request.content.read()
        bucket.body = body

        # Decode data from request body
        if body:

            # Decode data from JSON format
            if content_type.startswith('application/json'):
                return json.loads(body)

            # Decode data from x-www-form-urlencoded format
            elif content_type.startswith('application/x-www-form-urlencoded'):
                # TODO: Honor charset when receiving "application/x-www-form-urlencoded; charset=utf-8"
                payload = parse_qs(body, 1)
                # TODO: Decapsulate multiple values of same reading into "{name}-{N}", where N=1...
                for key, value in payload.iteritems():
                    if type(value) is types.ListType:
                        payload[key] = value[0]
                return payload

            # Decode data from CSV format
            elif content_type.startswith('text/csv'):

                if not self.metastore:
                    log.error('Generic decoding of CSV format requires metastore')

                # Prepare alias for metastore table
                csv_header_store = self.metastore.kotori['channel-csv-headers']
                #print 'csv_header_store:', csv_header_store

                # 1. Decode CSV header like '## weight,temperature, humidity' and remember for upcoming data readings
                def parse_header(channel_info, data_lines):

                    first_line = data_lines[0]
                    header_line = None
                    options = {}

                    # Regular header announcement
                    if first_line.startswith('## '):
                        header_line = first_line[3:].strip()
                        data_lines.pop(0)

                    # Convenience hack to support Open Hive CSV import
                    elif first_line.startswith('Date/Time') or first_line.startswith('Datum/Zeit'):
                        header_line = first_line
                        data_lines.pop(0)

                    # Convenience hack to support Beelogger CSV import
                    elif first_line.startswith('Datum,Uhrzeit'):
                        header_line = first_line
                        data_lines.pop(0)
                        options['rules'] = [{'type': 'fuse', 'source': ['Datum', 'Uhrzeit'], 'target': 'time', 'join': 'T', 'suffix': 'Z'}]

                    # Convenience hack to support import from http://archive.luftdaten.info/
                    elif first_line.startswith('sensor_id'):
                        header_line = first_line
                        data_lines.pop(0)

                    if header_line:
                        # Streamline various differences for even more convenience
                        header_line = header_line.replace(';', ',')
                        date_fields = ['Date/Time', 'Date', 'Datum/Zeit', 'timestamp']
                        for date_field in date_fields:
                            header_line = header_line.replace(date_field, 'time')
                        header_fields = map(str.strip, header_line.split(','))
                        msg = u'CSV Header: fields={fields}, key={key}'.format(fields=header_fields, key=request.channel_identifier)
                        log.info(msg)

                        csv_header_store.update_one(
                            {"channel": request.channel_identifier},
                            {"$set": {"header_fields": header_fields, "options": options}}, upsert=True)

                        message = u'Received header fields {}'.format(header_fields)
                        request.messages.append({'type': 'info', 'message': message})

                        channel_info['header_fields'] = header_fields
                        channel_info['options'] = options

                    #print 'header_fields, data_lines:', header_fields, data_lines
                    #return header_fields, data_lines

                # 2. Decode data, map to full-qualified payload container
                def parse_data(channel_info):
                    channel_info = channel_info or {}
                    data_raw = body.strip()
                    data_lines = map(str.strip, data_raw.split('\n'))
                    parse_header(channel_info, data_lines)
                    header_fields = channel_info.get('header_fields')
                    if not header_fields:
                        raise Error(http.BAD_REQUEST,
                            response='Could not process data, please supply field names via CSV header before sending readings')

                    #print 'data_lines:', data_lines; pprint(data_lines)

                    data_list = []
                    for data_line in data_lines:
                        data_fields = map(str.strip, data_line.replace(';', ',').split(','))
                        #print 'header_fields, data_fields:', header_fields, data_fields
                        data = OrderedDict(zip(header_fields, data_fields))
                        self.manipulate_data(data, channel_info)
                        data_list.append(data)

                    return data_list

                try:
                    channel_info = csv_header_store.find_one(filter={"channel": request.channel_identifier})
                except Exception as ex:
                    log.failure('Could not process CSV data, unknown database error: {0}'.format(ex))
                    raise Error(http.INTERNAL_SERVER_ERROR,
                        response='Could not process CSV data, unknown database error: {0}'.format(ex))

                return parse_data(channel_info)

            else:
                msg = u"Unable to handle Content-Type '{content_type}'".format(content_type=content_type)
                log.warn(msg)
                raise Error(http.UNSUPPORTED_MEDIA_TYPE, response=msg)

        else:
            msg = u'Empty request body'
            log.warn(msg)
            raise Error(http.BAD_REQUEST, response=msg)

    def manipulate_data(self, data, channel_info):
        """
        Data fusion on CSV data lines.
        Convenience hack to support Beelogger CSV import.
        """
        if 'options' in channel_info:
            rules = channel_info['options'].get('rules', [])
            for rule in rules:
                if rule['type'] == 'fuse':
                    extracted = []
                    for source_field in rule['source']:
                        if source_field in data:
                            extracted.append(data.get(source_field, ''))
                            del data[source_field]

                    separator = rule.get('join', '')
                    fused = separator.join(extracted)
                    fused += rule.get('suffix', '')

                    target = rule['target']
                    data[target] = fused


    def process_data(self, data, bucket):

        # Main transformation data container
        bucket.tdata = Bunch()

        # Merge request parameters (GET and POST) and url matches, in this order
        bucket.tdata.update(flatten_request_args(bucket.request.args))
        bucket.tdata.update(self.options.match)

        # When receiving a firmware request, run forwarding callback immediately
        is_firmware_request = 'slot' in bucket.tdata and bucket.tdata.slot == 'firmware'
        if is_firmware_request:
            bucket.data = data
            return self.options.callback(bucket)

        # Without any data received, run forwarding callback immediately
        if data is None:
            return self.options.callback(bucket)

        if type(data) is not types.ListType:
            data = [data]

        for item in data:
            # TODO: Apply this to telemetry values only!
            # FIXME: This is a hack
            if not bucket.request.path.endswith('event') and not bucket.request.path.endswith('firmware'):
                convert_floats(item, integers=['time'])
            self.propagate_single(item, bucket)

        message = 'Received #{number} readings'.format(number=len(data))
        bucket.request.messages.append({'type': 'info', 'message': message})


    def propagate_single(self, item, bucket):

        # Serialize as json for convenience
        # TODO: Really?
        item_json = json.dumps(item)

        # Update main bucket container
        bucket.data = item
        bucket.json = item_json

        # Run forwarding callback
        return self.options.callback(bucket)

    @staticmethod
    def error_response(bucket, error_message='', code=http.BAD_REQUEST, with_traceback=False):
        """
        Error handling method logging and returning appropriate stacktrace.
        """
        # FIXME: Check for privacy. Do something more sane with the stacktrace
        #        or enable only when sending appropriate request arguments.
        if with_traceback:
            error_message += '\n' + last_error_and_traceback()
            log.error(error_message)
        bucket.request.setResponseCode(code)
        #bucket.request.setHeader('Content-Type', 'text/plain; charset=utf-8')
        return error_message.encode('utf-8')


class HttpDataFrameResponse(object):
    """
    Bottom layer of data forwarding workhorse for HTTP.

    Generate appropriate output content based on
    information in transformation data ``bucket.tdata``.

    Render pandas DataFrame to various tabular and hierarchical
    data formats and different timeseries plots.

    Tabular data:

        - CSV
        - JSON
        - HTML
        - Excel (XLSX)
        - DataTables HTML widget

    Hierarchical data:

        - HDF5
        - NetCDF

    Timeseries plots:

        - [PNG]  matplotlib
        - [PNG]  ggplot
        - [HTML] dygraphs
        - [HTML] Bokeh
        - [HTML] Vega/Vincent

    """

    def __init__(self, bucket, dataframe):
        self.bucket = bucket
        self.request = bucket.request
        self.dataframe = dataframe

    def render(self):
        """
        Evaluate ``bucket`` information and enrich further before
        executing the designated output format rendering handler.
        """

        # Variable aliases
        bucket = self.bucket
        df = self.dataframe

        # Read designated suffix from transformation data
        suffix = bucket.tdata.suffix.lower()

        # Update "time_begin" and "time_end" fields to be in ISO 8601 format
        tdata = deepcopy(bucket.tdata)
        tdata.update({
            'time_begin': slugify_datettime(bucket.tdata.time_begin),
            'time_end':   slugify_datettime(bucket.tdata.time_end),
        })

        # Compute some names and titles and pluck into ``bucket``
        bucket.title = Bunch(
            compact = u'{gateway}_{node}'.format(**dict(tdata)).replace('-', '_'),
            short = u'{network}_{gateway}_{node}'.format(**dict(tdata)).replace('-', '_'),
            full  = u'{network}_{gateway}_{node}_{time_begin}-{time_end}'.format(**dict(tdata)).replace('-', '_'),
            human = u'Address: {network} » {gateway} » {node}'.format(**dict(tdata)),
        )


        # Buffer object most output handlers write their content to
        buffer = BytesIO()
        charset = None


        # Dispatch to appropriate output handler
        # TODO: XML, SQL, GBQ (Google BigQuery table), MsgPack?, Thrift?
        # TODO: jsonline using Odo, see http://odo.pydata.org/en/latest/json.html
        # TODO: Refactor "if response: return response" cruft
        # TODO: Refactor dispatching logic to improve suffix comparison redundancy with UniversalTabularExporter

        if suffix in ['csv', 'txt']:
            # http://pandas.pydata.org/pandas-docs/stable/io.html#io-store-in-csv
            df.to_csv(buffer, header=True, index=False, encoding='utf-8', date_format='%Y-%m-%dT%H:%M:%S.%fZ')
            charset = 'utf-8'

        elif suffix == 'tsv':
            df.to_csv(buffer, header=True, index=False, encoding='utf-8', date_format='%Y-%m-%dT%H:%M:%S.%fZ', sep='\t')
            charset = 'utf-8'

        elif suffix == 'json':
            # http://pandas.pydata.org/pandas-docs/stable/io.html#io-json-writer
            df.to_json(buffer, orient='records', date_format='iso')
            charset = 'utf-8'

        elif suffix == 'html':
            # http://pandas.pydata.org/pandas-docs/stable/io.html#io-html
            buffer.write('<html>\n')
            #buffer.write('<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">\n')
            df.to_html(buffer, index=False, justify='center')
            buffer.write('\n</html>')
            charset = 'utf-8'

        elif suffix == 'xlsx':
            exporter = UniversalTabularExporter(bucket, dataframe=df)
            response = exporter.render(suffix, buffer=buffer)
            if response:
                return response

        elif suffix in ['hdf', 'hdf5', 'h5']:
            exporter = UniversalTabularExporter(bucket, dataframe=df)
            response = exporter.render(suffix, buffer=buffer)
            if response:
                return response

        elif suffix in ['nc', 'cdf']:
            exporter = UniversalTabularExporter(bucket, dataframe=df)
            response = exporter.render(suffix, buffer=buffer)
            if response:
                return response

        elif suffix in ['dy', 'dygraphs']:
            plotter = UniversalPlotter(bucket, dataframe=df)
            response = plotter.render('html', kind='dygraphs')
            if response:
                return response

        elif suffix in ['dt', 'datatables']:
            exporter = UniversalTabularExporter(bucket, dataframe=df)
            response = exporter.render(suffix, buffer=buffer)
            if response:
                return response

        elif suffix in ['bk', 'bokeh']:
            plotter = UniversalPlotter(bucket, dataframe=df)
            response = plotter.render('html', kind='bokeh')
            if response:
                return response

        elif suffix == 'vega.json':
            plotter = UniversalPlotter(bucket, dataframe=df)
            response = plotter.render('json', kind='vega')
            if response:
                return response

        elif suffix == 'vega':
            plotter = UniversalPlotter(bucket, dataframe=df)
            response = plotter.render('html', kind='vega')
            if response:
                return response

        elif suffix in ['png']:
            plotter = UniversalPlotter(bucket, dataframe=df)
            response = plotter.render('png', buffer=buffer)
            if response:
                return response

        else:
            error_message = u'# Unknown data format "{suffix}"'.format(suffix=suffix)
            bucket.request.setResponseCode(http.BAD_REQUEST)
            bucket.request.setHeader('Content-Type', 'text/plain; charset=utf-8')
            return error_message.encode('utf-8')


        # Get hold of buffer content
        payload = buffer.getvalue()


        # Compute filename offered to browser
        filename = '{name}.{suffix}'.format(name=bucket.title.full, suffix=suffix)
        mimetype, encoding = mimetypes.guess_type(filename, strict=False)
        log.info(u'Fetching data succeeded, filename: {filename}, Format: {mimetype}', filename=filename, mimetype=mimetype)

        # Set "Content-Type" header
        if mimetype:
            content_type = mimetype
            if charset:
                content_type += '; charset=' + charset
            bucket.request.setHeader('Content-Type', content_type)

        # Set "Content-Disposition" header
        disposition = 'attachment'
        if mimetype in ['text/plain', 'text/csv', 'text/html', 'application/json', 'image/png']:
            disposition = 'inline'
        bucket.request.setHeader('Content-Disposition', '{disposition}; filename={filename}'.format(
            disposition=disposition, filename=filename))

        # Optionally encode to UTF-8 when serving HTML
        if mimetype == 'text/html':
            payload = payload.encode('utf-8')

        return payload




# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
import os
import re
import sys
import logging
import StringIO
import uuid
import arrow
from bunch import Bunch
from appdirs import user_cache_dir
from tempfile import NamedTemporaryFile
from contextlib import contextmanager
from pprint import pprint, pformat
from cornice.util import to_list
from git import Repo
from git import RemoteProgress
from plumbum import local
from plumbum.cmd import pwd, make, grep, ls
from kotori.util.errors import last_error_and_traceback

log = logging.getLogger(__name__)

# http://gitpython.readthedocs.io/en/stable/tutorial.html
# https://shell.readthedocs.io/en/latest/tutorial.html

"""
Todo
====

Prio 1
------
- [x] Return stderr and Python traceback in case of error
- [x] Finalize http pathway, incl. success (Content-Disposition) and failure (text/plain)
- [x] Run from system cache dir
- [x] Running "make" and "make firmware-info" consecutively takes too much time!
      Refactor into make target which has "all" as dependency target
- [x] Put "BOARD_TAG" and "BOARD_SUB" values from Makefile into "self.build_result"
      or maybe just pull them into the transformation dict, also with build path (target_path)
- [x] Don't forget to re-enable "origin.fetch" and "self.repo.submodule_update"

Prio 2
------
- [o] Option "prune" for pruning the buildpath before building
- [o] Option "clean" for running "make clean" before building
- [o] Add more logging to different "acquire_source" steps
- [o] Option for running "avr-strip" before handing out the ELF file
- [o] Add timing information to "self.build_result"
- [o] Propagate some information from "self.build_result" to HTTP response headers
- [o] Also redirect logging output to "self.stream"

Prio 3
------
- [o] Make configurable from which directory to run from to improve concurrent building (e.g. using a token)
- [o] Option "ansi=false" for disabling "CXXFLAGS += -fdiagnostics-color" in Makefile
- [o] Option for packaging the whole source tree in its state when building the software into a source tarball,
  including output from "git diff"
- [o] Make more generic by injecting Makefile
- [o] Make even more generic by making it handle arbitrary regular repositories of Arduino code
- [o] Bind gitlab/github webhook to "prune" operation
"""

class ProgressPrinter(RemoteProgress):
    def update(self, op_code, cur_count, max_count=None, message=''):

        if not message:
            return

        parts = (op_code, cur_count, max_count, cur_count / (max_count or 100.0), message or u"NO MESSAGE")
        parts = map(unicode, parts)
        message = u' '.join(parts)
        padding = u' ' * 120

        # STDERR
        #sys.stderr.write(message.strip() + padding)
        #sys.stderr.write('\r')
        #sys.stderr.write('\n')

        # Logger
        log.info(message)

class FirmwareBuilder(object):

    def __init__(self, repo_url=None, repo_branch='master', update_submodules=True, workingdir='.', architecture=None, esp_root=None):
        self.repo = None
        self.repo_url = repo_url
        self.repo_branch = repo_branch
        self.update_submodules = update_submodules
        self.workingdir = workingdir
        self.architecture = architecture or 'avr'
        self.esp_root = esp_root
        self.stream = sys.stderr

        self.repo_info = {}
        self.build_result = {}

        self.variable_patterns = [

            # Preprocessor defines .ino, .pde, .cpp, .h, e.g.
            # #define HE_USER         "testdrive"
            '#define\s+(?P<name>{name})\s+(?P<value>.+)',

            # Makefile variables, e.g.
            # BOARD_TAG         = pro328
            '^(?!#)(?P<name>{name})\s+:?=\s+(?P<value>.*)$',
            ]

    def acquire_source(self):

        cache_dir = os.path.join(user_cache_dir('kotori'), 'firmware-builder', 'sources')
        if not os.path.isdir(cache_dir): os.makedirs(cache_dir)

        target_path = os.path.join(cache_dir, os.path.basename(self.repo_url))

        log.info('Build path is {target_path}'.format(target_path=target_path))

        # 1. Get hold of repository
        if os.path.isdir(target_path):

            # Define repository object from existing path
            # http://gitpython.readthedocs.io/en/stable/tutorial.html#meet-the-repo-type
            self.repo = Repo(target_path)

            # Fetch updates from remote
            # http://gitpython.readthedocs.io/en/stable/tutorial.html#handling-remotes
            origin = self.repo.remotes.origin
            log.info('Fetching updates from repository {repository}'.format(repository=origin.url))
            origin.fetch(progress=ProgressPrinter())

        else:

            # Clone repository from a remote
            log.info('Cloning repository {repository}'.format(repository=self.repo_url))
            self.repo = Repo.clone_from(self.repo_url, target_path, progress=ProgressPrinter())


        # Default effective gitref to given reference label (either branch/tag name or commit sha)
        gitref_effective = self.repo_branch

        # Compute effective gitref by searching the origin branches for the given reference label
        origin = self.repo.remotes.origin
        origin_branch_ref = origin.name + '/' + self.repo_branch
        for ref in origin.refs:
            if ref.name == origin_branch_ref:
                gitref_effective = ref.name


        # 2. Switch to designated branch/gitref
        # http://gitpython.readthedocs.io/en/stable/tutorial.html#switching-branches

        # Point the HEAD symbolic reference to the designated gitref
        self.repo.head.reference = gitref_effective

        # Brutally reset the index and working tree to match the pointed-to commit
        # Attention: Don't do this on a real repository you are working on, data may get lost!
        self.repo.head.reset(index=True, working_tree=True)


        # 3. Initialize submodules
        # http://gitpython.readthedocs.io/en/stable/tutorial.html#advanced-repo-usage
        # update all submodules, non-recursively to save time, this method is very powerful, go have a look
        if self.update_submodules:
            log.info('Updating git submodules')
            self.repo.submodule_update(recursive=True, progress=ProgressPrinter())
        else:
            log.info('Skip updating git submodules')

        # Return some information about repository
        self.repo_info = {
            'remote': origin.url,
            'ref': gitref_effective,
            'commit': self.repo.head.commit.hexsha[:8],
        }

    def patch_files(self, filepatterns, data):

        filepatterns = to_list(filepatterns)

        directory = os.path.join(self.repo.working_dir, self.workingdir)

        # cd into appropriate directory
        with local.cwd(directory):
            for pattern in filepatterns:

                # Find files matching pattern in directory
                # This is plumbum's syntax for shell globbing
                path = local.path()
                lp_files = path // str(pattern)

                # Iterate and patch all files
                for lp_file in lp_files:
                    filepath = str(lp_file)
                    self.patch_file(filepath, data)

    def patch_file(self, filepath, data):

        log.info('Patching file {filepath}'.format(**locals()))

        # Read file
        payload = file(filepath, 'r').read()

        # Machinery for computing replacements
        replacements = []
        for name, value in data.iteritems():
            for match in self.find_variable(payload, name, multiline=True):

                value_before = match['value']

                is_quoted = value_before.startswith('"')

                # Debugging: Provoke build error when replacing e.g. GPRSBEE_AP_NAME
                #is_quoted = value_before.startswith('"') and value_before.endswith('"')

                replacement = str(value)
                if is_quoted:
                    replacement = '"' + replacement + '"'
                match['line-after'] = match['line-before'].replace(value_before, replacement)
                replacements.append(match)

        log.info('Replacements:\n%s', pformat(replacements))

        # Apply replacements
        for replacement in replacements:
            payload = payload.replace(replacement['line-before'], replacement['line-after'])

        # Write file
        file(filepath, 'w').write(payload)

    def find_variable(self, payload, name, multiline=False):
        for pattern_template in self.variable_patterns:
            pattern = pattern_template.format(name=name)
            if multiline:
                matcher = re.compile(pattern, re.MULTILINE)
            else:
                matcher = re.compile(pattern)
            for m in matcher.finditer(payload):
                match = m.groupdict()
                match['line-before'] = m.group(0)
                # Exit inner loop
                break
            else:
                # Cycle outer loop when having no matches
                continue

            yield match

    def run_build(self, makefile=None):
        log.info('Running build')
        if self.architecture.lower() == 'esp':
            with local.env(ESP_ROOT=self.esp_root):
                return self.run_makefile(makefile=makefile)
        else:
            return self.run_makefile(makefile=makefile)

    def run_makefile(self, makefile=None):
        """
        Run the whole build process with designated Makefile.
        """

        # cd into git repository directory
        with local.cwd(self.repo.working_dir):

            # cd into working directory inside git repository
            with local.cwd(self.workingdir):

                # Run Makefile to start the compilation process
                make('--file', makefile, 'all-plus-firmware-info', stdout=self.stream, stderr=self.stream)

                # Slurp output of build process
                try:
                    self.stream.seek(0)
                    output = self.stream.read()
                except IOError:
                    make_firmware_info = make['--file', makefile, 'firmware-info'] | grep['TARGET_']
                    output = make_firmware_info()

                # Grep "TARGET_HEX", "TARGET_ELF" (for AVR) as well as "TARGET_BIN" and "TARGET_CHIP" (for ESP) paths
                # from build output and store into "self.build_result"
                target_matcher = re.compile('(?P<name>TARGET_.+):(?: (?P<value>.+))?')
                for m in target_matcher.finditer(output):
                    match = m.groupdict()
                    name  = match['name']
                    value = match['value']
                    if value:
                        self.build_result[name] = value

                # Add build path to build result
                self.build_result['build_path'] = pwd().strip()

    def make_artefact(self):
        artefact = Artefact()

        artefact.architecture = self.architecture
        artefact.source = self.repo_info
        artefact.build  = self.build_result

        # AVR
        if self.architecture == 'avr':
            target_hex = os.path.abspath(os.path.join(self.build_result['build_path'], self.build_result['TARGET_HEX']))
            target_elf = os.path.abspath(os.path.join(self.build_result['build_path'], self.build_result['TARGET_ELF']))

            artefact.name = os.path.splitext(os.path.basename(target_hex))[0]
            artefact.binary.hex = file(target_hex, 'rb').read()
            artefact.binary.elf = file(target_elf, 'rb').read()

        # ESP
        elif self.architecture == 'esp':
            target_bin = os.path.abspath(self.build_result['TARGET_BIN'])
            artefact.name = os.path.splitext(os.path.basename(target_bin))[0]
            artefact.binary.bin = file(target_bin, 'rb').read()

        return artefact

    @contextmanager
    def capture(self):
        """
        Capture output of the whole build process.
        Return stacktrace of Python driver and output of make command, if any.
        """

        # The result object containing build success flag and error outputs.
        # This gets bound to the context manager variable.
        result = Bunch()
        self.build_result['capture'] = result

        # A temporary file to redirect make output to
        self.stream = NamedTemporaryFile()
        try:
            # Pass execution flow to context manager body,
            # effectively running the main build process
            yield result

            # Signal success if build engine completed
            result.success = True

        except Exception as ex:

            # Signal failure
            result.success = False

            # Capture Python traceback
            result.error = last_error_and_traceback()

        # Capture make output
        self.stream.seek(0)
        result.output = self.stream.read()

    def error_message(self):

        result = self.build_result['capture']
        buf = StringIO.StringIO()
        width = 80

        print >>buf, u'=' * width
        print >>buf, u'INFO'.center(width)
        print >>buf, u'=' * width
        print >>buf
        print >>buf, 'Repository:'
        print >>buf, pformat(self.repo_info, indent=12)
        print >>buf
        print >>buf, 'Directory: ', self.workingdir
        print >>buf

        print >>buf, u'=' * width
        print >>buf, u'ERROR'.center(width)
        print >>buf, u'=' * width
        print >>buf
        print >>buf, result.error.decode('utf-8')
        print >>buf

        print >>buf, u'=' * width
        print >>buf, u'OUTPUT'.center(width)
        print >>buf, u'=' * width
        print >>buf
        print >>buf, result.output.decode('utf-8')

        return buf.getvalue()

class Artefact(object):

    def __init__(self):

        # TODO: Add more build arguments
        # TODO: Persist somehow
        # TODO: Display contents on html page

        self.architecture = None
        self.source  = None
        self.build   = None
        self.name    = None

        self.binary  = BinaryArtefact()

        self.identifier = str(uuid.uuid4())
        self.timestamp  = arrow.now().format('YYYYMMDDTHHmmss')

    def get_binary(self, suffix):
        return getattr(self.binary, suffix)

    @property
    def commit_sha(self):
        return self.source['commit']

    @property
    def fullname(self):

        template = '{name}_{architecture}-{target}_{timestamp}_{commit}_{identifier}'

        target = 'unknown'
        commit = self.commit_sha

        if self.architecture == 'avr':
            target = '{board_tag}-{board_sub}'.format(
                board_tag = self.build.get('TARGET_BOARD_TAG', 'unknown'),
                board_sub = self.build.get('TARGET_BOARD_SUB', 'unknown')
            )

        elif self.architecture == 'esp':
            target = '{target_chip}'.format(
                target_chip = self.build.get('TARGET_CHIP', 'unknown'),
            )

        variables = {}
        variables.update(**self.__dict__)
        variables.update(**locals())

        result = template.format(**variables)

        return result

    def __repr__(self):
        hex_size = self.binary.hex and len(self.binary.hex) or None
        elf_size = self.binary.elf and len(self.binary.elf) or None
        bin_size = self.binary.bin and len(self.binary.bin) or None
        tplvars = {}
        tplvars.update(**self.__dict__)
        tplvars.update(**locals())
        return u'<{class_} name={name} source={source} hexsize={hex_size} elfsize={elf_size} binsize={bin_size}>'.format(class_=type(self).__name__, **tplvars)


class BinaryArtefact:

    def __init__(self):

        self.hex     = None
        self.elf     = None
        self.bin     = None



# ------------------------------------------
#   Example program
# ------------------------------------------

def example_recipe(fb):

    fb.acquire_source()

    data = {
        'BOARD_TAG': 'mega2560',
        'BOARD_SUB': 'atmega2560',
        'HE_USER': 'Hotzenplotz',
        'HE_SITE': 'Buxtehude',
        'HE_HIVE': 'Raeuberhoehle',
        'GPRSBEE_APN': 'internet.altes-land.de',
        'GPRSBEE_APNUSER': 'kaschperl',
        'GPRSBEE_APNPASS': '12345',
        'GPRSBEE_VCC': 23,
        }
    fb.patch_files(['*.ino', '*.pde', '*.cpp', '*.h', 'Makefile*'], data)

    fb.run_makefile(makefile='Makefile-OSX.mk')
    #fb.run_makefile(makefile='Makefile-FWB.mk')
    artefact = fb.make_artefact()
    #print artefact.elf

    return artefact

def firmware_builder():
    repo_url = 'https://github.com/hiveeyes/arduino'
    return FirmwareBuilder(repo_url=repo_url, repo_branch='master', workingdir='node-gprs-any')

def run_example():
    fb = firmware_builder()
    artefact = example_recipe(fb)
    log.info(u'Build succeeded, artefact: {}'.format(artefact))

def capture_example():
    fb = firmware_builder()
    with fb.capture() as result:
        artefact = example_recipe(fb)

    print fb.build_result.keys()

    if result.success:
        log.info(u'Build succeeded, artefact: {}'.format(artefact))
    else:
        log.error(u'Build failed\n{message}'.format(message=fb.error_message()))

    return result


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
    #run_example()
    result = capture_example()
    print result.output




# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl <andreas@getkotori.org>
import re


class MessageType:

    DATA_DISCRETE = 1
    DATA_CONTAINER = 2
    EVENT = 3
    ERROR = 3


class TopicPatterns:

    data = [

        # Legacy v1 (deprecated)
        'message-json',

        # En bloc
        '/data.json',

        # Homie
        'data/__json__',

        # WeeWX
        'loop',

        # Sonoff-Tasmota
        '/SENSOR',
        '/STATE',
    ]

    discrete = [
        # Discrete value
        # {base}/data/temperature
        '/(data|measure)/[^/]+',
    ]

    event = [

        # Record of event data
        'event.json',

        # Homie support
        'event/__json__',
    ]


    error = [

        # Error data arrives on this suffix
        'error.json',
    ]

    @classmethod
    def compile_patterns(cls, patterns):
        matcher = re.compile('({parts})'.format(parts='|'.join(map(lambda x: x + '$', patterns))))
        return matcher


class TopicMatchers:
    data = TopicPatterns.compile_patterns(TopicPatterns.data)
    discrete = TopicPatterns.compile_patterns(TopicPatterns.discrete)
    event = TopicPatterns.compile_patterns(TopicPatterns.event)
    error = TopicPatterns.compile_patterns(TopicPatterns.error)
